/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE	600

#include "ssl-common.h"

#include <stdlib.h>
#include <limits.h>
#include <math.h>

#include <sys/select.h>
#include <sys/types.h>
#include <pwd.h>

#include <unistd.h>

#include <string.h>
#include <strings.h>

#ifdef __APPLE__
/* For gettimeofday() */
# include <sys/time.h>
#else	/* __APPLE__ */
/* For clock_gettime() */
# include <time.h>
# ifdef CLOCK_MONOTONIC_RAW
#  define XACML_IO_CLOCK_ID  CLOCK_MONOTONIC_RAW
# else	/* CLOCK_MONOTONIC_RAW */
#  define XACML_IO_CLOCK_ID  CLOCK_MONOTONIC
# endif /* CLOCK_MONOTONIC_RAW */
#endif	/* __APPLE__ */

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>
#include <openssl/rand.h>

#include "xacml_io_ssl_log.h"

#define	SERIALSTR_LENGTH    ((size_t)255L)

/* Maximum total, re-read timeout for SSL_read() in xacml_io_SSL_recv()
 * given in seconds */
#define SSL_READ_TIMEOUT    30L


/* SSL_COMP_get_compression_methods() has been in OpenSSL since version 0.9.2
 * but the prototype was missing before 0.9.8 */
#if OPENSSL_VERSION_NUMBER < 0x00908000L
STACK_OF(SSL_COMP) *SSL_COMP_get_compression_methods(void);
#endif


/*************************************************************************
 * Global variables, defined extern in ssl-common.h
 *************************************************************************/

SSL_CTX * xacml_io_ssl_ctx    = NULL;
char * xacml_io_certCAPath    = NULL;
char * xacml_io_certFilePub   = NULL;
char * xacml_io_certFilePriv  = NULL;
char * xacml_io_certOwner     = NULL;

/*************************************************************************
 * Static global variables
 *************************************************************************/

static char * ssl_override_hostname = NULL;

/* Is OpenSSL already initialized */
static int openssl_is_initialized = 0;

#ifndef DISABLE_PROXY_SUPPORT
/* Default is to enable proxy support */
static int disable_proxy_support = 0;
#endif /* DISABLE_PROXY_SUPPORT */

/******************************************************************************
 * Private functions
 ******************************************************************************/
static int changeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid);
static int restoreEffectiveToRealUid (uid_t eff_uid);

#ifndef DISABLE_PROXY_SUPPORT
static int grid_X509_knownCriticalExts(X509 *cert);
static int grid_verifyProxy( STACK_OF(X509) *certstack );
static time_t grid_asn1TimeToTimeT(char *asn1time);
static int scas_verify_callback(int ok, X509_STORE_CTX *store);
static int grid_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer);
static void grid_get_serialStr(ASN1_INTEGER *serial);
static time_t my_timegm(struct tm *tm);
#endif /* DISABLE_PROXY_SUPPORT */

/******************************************************************************
Function:     xacml_io_SSL_recv
Description:
              This function will use the SSL object to readout the SSL transport
              receive buffer for num amount of bytes and put it to the data buffer.
              Error message handling included
              Note the entire SSL read buffer will be read, until it doesn't fit
              in the buffer.
Parameters:
Returns:
     < 0  :  Error code from SSL_read()
    == 0  :  Nothing to be done
     > 0  :  Amount of bytes read from the buffer
******************************************************************************/
int xacml_io_SSL_recv (SSL * ssl, void * data, int num)
{
    int error_code = 0;
    int rc = 0;
#ifdef __APPLE__
    struct timeval   t1,t2;
#else
    struct timespec  t1,t2;
#endif
#if 0
    int nread = 0;
    unsigned int retry = 0;
#endif

    XACML_LOG_DEBUG(5, "%s: entering %s\n", __func__, __func__);

    if (!ssl)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: No SSL object presented to xacml_io_SSL_recv()\n", __func__);
        return -1;
    }

    /* clear buffer, note that num is an int as this is what is needed in
     * ssl_io_recv which is used in a xacml_io_descriptor_t, see xacml_io_ssl.c
     * and ultimately by gsoap. * */
    memset(data, '\0', (size_t)num);

    /* rc is the return code when rc <= 0, otherwise its the amount of bytes read */

    /*
     * Igor's patch to be more resiliant to long distance connections
     * which don't fill the buffers fast enough to let this work at once
     *
    rc = SSL_read (ssl, data, num);
    if (rc < 0)
        xacml_io_print_ssl_error_msg (ssl, rc, 1, &error_code);
    */

    /* Note: MacOS does not have clock_gettime() */
#ifdef __APPLE__
    if (gettimeofday(&t1,NULL))
#else
    if (clock_gettime(XACML_IO_CLOCK_ID,&t1))
#endif
    {
	/* this can only happen when something is really wrong... */
        XACML_LOG_FUNC(LOG_ERR, "%s: cannot get time of day: %s\n",
		__func__, strerror(errno));
	return -1; /*  */
    }

/* Endless retry (this might bite me) */
/*    while (1)*/
    do
    {
        rc = SSL_read (ssl, data, num);
	if (rc>0)   {
	    XACML_LOG_DEBUG(1, "%s: Read %d bytes, buffer size is %d.\n", __func__, rc, num);
	    return rc;
        }

	/* rc<=0, which for SSL_read can also return 0 on error */
	if (xacml_io_print_ssl_error_msg (ssl, rc, 1, &error_code))
	    return rc; /* fatal error */

	/* recoverable: retry */
#ifdef __APPLE__
	/* Note: MacOS does not have clock_gettime() */
	if (gettimeofday(&t2,NULL))
#else
	if (clock_gettime(XACML_IO_CLOCK_ID,&t2))
#endif
	{
	    /* this can only happen when something is really wrong... */
	    XACML_LOG_FUNC(LOG_ERR, "%s: cannot get time of day: %s\n",
				    __func__, strerror(errno));
	    return -1; /* cannot check what time it is... */
	}
    }
#ifdef __APPLE__
    while ( (double)(t2.tv_sec-t1.tv_sec)+((double)(t2.tv_usec-t1.tv_usec))/1e6 < SSL_READ_TIMEOUT);
#else
    while ( (double)(t2.tv_sec-t1.tv_sec)+((double)(t2.tv_nsec-t1.tv_nsec))/1e9 < SSL_READ_TIMEOUT);
#endif

    /* We had a recoverable error but hit the timeout */
    XACML_LOG_FUNC(LOG_WARNING,"%s: Hit maximum re-read timeout of %ld seconds.\n",
	    __func__, SSL_READ_TIMEOUT);

    return rc;

#if 0
  /* Out of scope */
    while ((nread < num) &&
           (rc = SSL_read (ssl, (unsigned char *) data + nread, num - nread)) &&
           (retry < 5))
    {
        /* Got bytes */
        if (rc > 0)
        {
            nread += rc;
        }
        else
        {
            if (xacml_io_print_ssl_error_msg (ssl, rc, 1, &error_code))
                break;
            else
            {
                retry++;
                XACML_LOG_DEBUG(3, "%s: continueing SSL_read(), retry %d\n", __func__, retry);
                usleep (50 * 1000 * retry);
                continue;
            }
        }
    }
    XACML_LOG_DEBUG(1, "%s: Read %d bytes, buffer size is %d.\n", __func__, nread, num);

    xacml_io_print_ssl_error_msg (ssl, rc, 0, &error_code);
    if ((error_code == SSL_ERROR_WANT_READ) || (error_code == SSL_ERROR_WANT_WRITE))
    {
        XACML_LOG_FUNC(LOG_NOTICE, "%s: Something is still in the transport buffer\n", __func__);
    }

    return nread ? nread : rc;
#endif
}



/******************************************************************************
Function:     xacml_io_SSL_server_accept
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect()
              handshake

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * xacml_io_SSL_server_accept (int opensocket, SSL_CTX * ctx)
{
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int              i     = 0;
    int              depth = 0;
    int              ssl_error_code = 0;
    X509 *           ssl_peer_cert       = NULL;
    STACK_OF(X509) * ssl_peer_cert_chain = NULL;
    char           * dn;
    int activity = 0;
    fd_set readfds;
    struct timeval timeout;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: Failed to create the SSL layer\n", __func__);
            return NULL;
        }
        else
        {
            XACML_LOG_DEBUG(5, "%s: SSL object created from CTX\n", __func__);
            SSL_clear(ssl);
        }
    }
    else
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: can't create SSL handle out of CTX structure\n", __func__);
        return NULL;
    }


    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Failed to tie the socket to a SSL BIO\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: BIO created from socket\n", __func__);
    }


    /* Tie the socket as a file descriptor to the SSL layer */
    if (! SSL_set_fd(ssl, opensocket))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: SSL_set_fd() failed!\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Passed setting binding fd to SSL\n", __func__);
    }


    /* Set the connection to non-blocking mode */
    /*
    BIO_set_nbio (bio, 1);
     */

    SSL_set_bio(ssl, bio, bio);
    XACML_LOG_DEBUG(5, "%s: BIO set for SSL\n", __func__);

    BIO_set_nbio(bio,1);
    XACML_LOG_DEBUG(5, "%s: BIO set to non-blocking\n", __func__);

/*
    SSL_set_accept_state(ssl);
    XACML_LOG_DEBUG(5, "%s: Passed SSL set accept state\n", __func__);
*/


/********************************************/
    timeout.tv_sec=0;
    timeout.tv_usec=500000;

    XACML_LOG_DEBUG(5, "%s: Waiting for client activity...\n", __func__);
    for (;;)
    {
        FD_ZERO(&readfds);

        /* setup which sockets to listen on */
        FD_SET(opensocket, &readfds);

        /* wait for connection, forever if we have to */
        activity = select(opensocket + 1, &readfds, (fd_set*)0, (fd_set*)0, &timeout);

        if ((activity < 0) && (errno!=EINTR))
        {
            /* there was an error with select() */
            SSL_free(ssl);
            return NULL;
        }
        else if (activity == 0)
        {
            /* Time out reached */
            XACML_LOG_DEBUG(3, "%s: Client SSL activity time out reached.\n", __func__);
            SSL_free(ssl);
            return NULL;
        }
        else
        {
            if (FD_ISSET(opensocket, &readfds))
            {
                XACML_LOG_DEBUG(5, "%s: Got client activity.\n", __func__);
                break;
            }
        }
    }

/********************************************/

    /* SSL accepting */
    XACML_LOG_DEBUG(5, "%s: Waiting in SSL_accept()\n", __func__);
    if ((rc = SSL_accept(ssl)) <= 0)
    {
        XACML_LOG_DEBUG(3, "%s: Warning: failure in SSL_accept()\n", __func__);
        if (xacml_io_print_ssl_error_msg (ssl, rc, 1, &ssl_error_code))
        {
            SSL_free(ssl);
            return NULL;
        }
        else
        {
            if ((ssl_error_code == SSL_ERROR_WANT_READ) || (ssl_error_code == SSL_ERROR_WANT_WRITE))
            {
                XACML_LOG_FUNC(LOG_NOTICE, "%s: Something is still in the transport buffer\n", __func__);
                SSL_free(ssl);
                return NULL;
            }

            /* SSL accepting - retry */
            if ((rc = SSL_accept(ssl)) <= 0)
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Error: failure in SSL_accept() (already retried)\n", __func__);
                xacml_io_print_ssl_error_msg (ssl, rc, 1, NULL);
                SSL_free(ssl);
                return NULL;
            }
        }
    }
    XACML_LOG_DEBUG(5, "%s: Successful SSL_accept()\n", __func__);

/***/

    if (!(ssl_peer_cert = SSL_get_peer_certificate(ssl)))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Could not get peer certificate from the SSL handle.\n", __func__);
        SSL_free(ssl);
    }

    if (!(ssl_peer_cert_chain = SSL_get_peer_cert_chain(ssl)))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
        SSL_free(ssl);
    }
    else
    {
        /*
         * So we have a chain... but does it contain the peer certificate?
         * ... lets check and add it if not
         */

        depth = sk_X509_num(ssl_peer_cert_chain);
        XACML_LOG_DEBUG(5, "%s: Chain depth = %d\n", __func__, depth);
        sk_X509_insert(ssl_peer_cert_chain, ssl_peer_cert, 0);
        XACML_LOG_DEBUG(5, "%s: Added the peer certificate to the chain\n", __func__);
        depth = sk_X509_num(ssl_peer_cert_chain);
        XACML_LOG_DEBUG(5, "%s: Chain depth = %d\n", __func__, depth);

        for (i = 0; i < depth; i++)
        {
            dn = X509_NAME_oneline(X509_get_subject_name (sk_X509_value(ssl_peer_cert_chain, i)), NULL, 0);

            XACML_LOG_DEBUG(5, "%s: (%d) -> \"%s\" (%s) .\n",
                               __func__,
                               i,
                               dn,
                               xacml_io_x509IsCA(sk_X509_value(ssl_peer_cert_chain, i)) ? "is CA" : "not a CA");

            free( dn );
        }
    }


    /* Connection stuff done, client connected */
    XACML_LOG_DEBUG(4, "%s: SSL Connection accepted and opened\n", __func__);

    return ssl;
}



/******************************************************************************
Function:     xacml_io_SSL_client_connect
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect()
              handshake

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * xacml_io_SSL_client_connect (int opensocket, SSL_CTX * ctx)
{
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int    retry = 0;
    unsigned long code;
    const char *reason;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
	    code=ERR_get_error();
	    if (code)   {
		reason=ERR_reason_error_string(code);
		if (!reason)
		    reason=ERR_error_string(code,NULL);
	    } else
		reason="unknown";
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: Failed to create the SSL layer. OpenSSL reason: %s.\n",
                        __func__,
			reason);
            return NULL;
        }
        else
        {
            XACML_LOG_DEBUG(5, "%s: SSL object created from CTX\n", __func__);
        }
    }
    else
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: can't create SSL handle out of CTX structure\n", __func__);
        return NULL;
    }



    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Failed to tie the socket to a SSL BIO\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: BIO created from socket\n", __func__);
    }


    /* Set the connection to non-blocking mode */
    /*
    BIO_set_nbio (bio, 1);
     */

    /* Binding/setting the BIO to the SSL layer */
    SSL_set_bio(ssl, bio, bio);
    XACML_LOG_DEBUG(5, "%s: BIO set for SSL\n", __func__);


    /* The retries are local to the system mostly, hence you can try a lot per connected client */
    if ((rc = SSL_connect(ssl)) <= 0)
    {
	/* Note: SSL_connect can end with -1 if it isn't finished yet, see
	 * manpage. Error will be printed in xacml_io_print_ssl_error_msg() */
	if (errno)
	    XACML_LOG_FUNC(LOG_DEBUG,
		"%s: SSL_connect() ended with rc=%d (%s)\n",
		__func__,rc,strerror(errno));
	else
	    XACML_LOG_FUNC(LOG_DEBUG,
		"%s: SSL_connect() ended with rc=%d\n",
		__func__,rc);
        if (xacml_io_print_ssl_error_msg (ssl, rc, 1, NULL))
        {
	    /* Unrecoverable */
            SSL_free(ssl);
            return NULL;
        }
        else
        {
            while (retry < 222)
            {
                if ((rc = SSL_connect(ssl)) <= 0)
                {
                    if (xacml_io_print_ssl_error_msg (ssl, rc, 0, NULL))
                    {
                        break;
                    }
                    else
                    {
			/* According to the SSL error, you should try again.
			 * Introducing an I/O micro sleep */
                        retry++;
                        continue;
                    }
                }
                else
                {
                    /* Success */
                    XACML_LOG_DEBUG(5,"%s: SSL Connection opened (retry=%d)\n",
			    __func__, retry);
                    return ssl;
                }
            }

            /* Non-succesfull / failure condition */
            /* If here, then the SSL connection isn't meant to be, so bailing out */
            xacml_io_print_ssl_error_msg (ssl, rc, 1, NULL);
	    XACML_LOG_FUNC(LOG_DEBUG, "%s: SSL_connect() failed after retry %d\n",
			__func__, retry);
            SSL_free(ssl);
            return NULL;
        }
    }
    else
    {
        /* Success */
        XACML_LOG_DEBUG(5, "%s: SSL Connection opened\n", __func__);
        return ssl;
    }
}


/******************************************************************************
Function:     xacml_io_SSL_connect_shutdown
Description:
              The SSL object will be shutdown fully

Parameters:
Returns:
        0 : Perfect shutdown
        1 : FUBAR
******************************************************************************/
int xacml_io_SSL_connect_shutdown (SSL * ssl)
{
    unsigned int retry = 0;
    int rc = 0;

    if (!ssl)
    {
        XACML_LOG_DEBUG(5, "%s: Shutting down SSL; nothing to shutdown though... empty SSL pointer presented.\n", __func__);
        return 1;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Shutting down SSL\n", __func__);
        while (retry < 10)
        {
          switch (rc = SSL_shutdown (ssl))
          {
            /* First SSL_shutdown must return 0 */
            /* Second SSL_shutdown must return 1 */
            case 0 :
                if (retry == 0)
                    XACML_LOG_DEBUG(5, "%s: First shutdown succeeded, close notify send ok\n", __func__);
                else
                    XACML_LOG_DEBUG(2, "%s: Retrying shutdown procedure (try: %d)\n", __func__, retry + 1);
		break;
            case 1 :
                XACML_LOG_DEBUG(2, "%s: Succesfull shutdown #1\n", __func__);

                XACML_LOG_DEBUG(5, "%s: Freeing SSL BIO's\n", __func__);
                BIO_free( SSL_get_rbio( ssl ) );

/*                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL);*/
		/* Should not call SSL_CTX_free() here, it will cause a SEGV
		 * since SSL_free also free some/all of that memory */
/*                SSL_CTX_free( SSL_get_SSL_CTX( ssl ) );*/

                XACML_LOG_DEBUG(2, "%s: Succesfull shutdown #2\n", __func__);
                return 0;
            default :
                if (xacml_io_print_ssl_error_msg (ssl, rc, 1, NULL)==0)	{
		    /* recoverable error */
		    XACML_LOG_FUNC(LOG_DEBUG,
			"%s: Shutdown not (yet) successful, retry count is: %d\n",
			__func__, retry + 1);
		    break;
		}
		/* Unrecoverable error */
		XACML_LOG_FUNC(LOG_NOTICE,
			"%s: Shutdown failed completely, retry count is: %d, SSL_get_shutdown()=%d\n",
			__func__, retry + 1, SSL_get_shutdown(ssl));

                XACML_LOG_DEBUG(5, "Freeing SSL BIO's\n");
                BIO_free( SSL_get_rbio( ssl ) );

/*                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL);*/
		/* Should not call SSL_CTX_free() here, it will cause a SEGV
		 * since SSL_free also free some/all of that memory */
/*                SSL_CTX_free( SSL_get_SSL_CTX( ssl ) );*/

                return 1;
	    }
	    retry++;

	    /* Perform microsleep to cut the system some slack to catch up */
	    usleep (500 + 500 * retry);
	    continue;
        }

	/* Too many retries */
	XACML_LOG_FUNC(LOG_NOTICE,
			    "%s: Too many retries (%d) trying to shutdown\n",
			    __func__, retry + 1);

	XACML_LOG_DEBUG(5, "Freeing SSL BIO's\n");
	BIO_free( SSL_get_rbio( ssl ) );
    }

    /* Shouldn't get here */
    return 1;
}




/******************************************************************************
Function:   xacml_io_setup_client_ctx
Description:
    Setting up the SSL client side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *xacml_io_setup_client_ctx(char * cafile,
				   char * cadir,
				   char * clientcert,
				   char * clientkey)
{
    SSL_CTX * ctx = NULL;
    int changed_id_failed=0;
    uid_t stored_real_uid=(uid_t)-1;
    uid_t stored_eff_uid=(uid_t)-1;
    unsigned long code;
    const char *reason;


    XACML_LOG_DEBUG(5, "Entering: %s\n", __func__);
    ctx = SSL_CTX_new(SSLv23_client_method());
    if (!ctx)
    {
	code=ERR_get_error();
	if (code)   {
	    reason=ERR_reason_error_string(code);
	    if (!reason)
		reason=ERR_error_string(code,NULL);
	} else
	    reason="unknown";
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: No SSL context was created. OpenSSL error: %s.\n",
                    __func__, reason);
        return NULL;
    }


    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
	code=ERR_get_error();
	if (code)   {
	    reason=ERR_reason_error_string(code);
	    if (!reason)
		reason=ERR_error_string(code,NULL);
	} else
	    reason="unknown";
	XACML_LOG_FUNC(LOG_ERR, "%s: Error loading CA file and/or directory: file: %s dir: %s. OpenSSL error: %s.\n",
		    __func__,
		    cafile ? cafile : "n/a",
		    cadir ? cadir : "n/a",
		    reason);
        SSL_CTX_free (ctx);
        return NULL;
    }


    /* Change account, by default to bypass NFS root squash problems */
    if ( (changed_id_failed=changeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid)) )
	XACML_LOG_FUNC(LOG_NOTICE,"%s: changeEffectiveToRealUid() failed.\n",__func__);
    if (SSL_CTX_use_certificate_chain_file(ctx, clientcert) != 1)
    {
	code=ERR_get_error();
	if (code)   {
	    reason=ERR_reason_error_string(code);
	    if (!reason)
		reason=ERR_error_string(code,NULL);
	} else
	    reason="unknown";
        XACML_LOG_FUNC(LOG_ERR, "%s: Error loading certificate chain from file: '%s'. OpenSSL error: %s.\n",
                    __func__,
                    clientcert ? clientcert : "n/a",
		    reason);
        SSL_CTX_free (ctx);
        return NULL;
    }

    if (SSL_CTX_use_PrivateKey_file(ctx, clientkey, SSL_FILETYPE_PEM) != 1)
    {
	code=ERR_get_error();
	if (code)   {
	    reason=ERR_reason_error_string(code);
	    if (!reason)
		reason=ERR_error_string(code,NULL);
	} else
	    reason="unknown";
        XACML_LOG_FUNC(LOG_ERR, "%s: Error loading private key from file: '%s'. OpenSSL reason: %s.\n",
                    __func__,
                    clientkey,
		    reason);
        SSL_CTX_free (ctx);
        return NULL;
    }

    /* Restore to the previous settings */
    if (!changed_id_failed) {
	if (restoreEffectiveToRealUid (stored_eff_uid))	{
	    XACML_LOG_FUNC(LOG_WARNING,"%s: Cannot restore effective uid to %d\n",
		    __func__,stored_eff_uid);
	}
    }

    /*
     * Go into verify_callback after default SSL checks
     * Perform signature checks (peer must present certificate and must verify it)
     */
#ifndef DISABLE_PROXY_SUPPORT
    if (disable_proxy_support==0)   {
	SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, scas_verify_callback);
	/* Probably needs to be here */
	ctx->cert_store->check_issued = grid_check_issued_wrapper;

# if OPENSSL_VERSION_NUMBER < 0x00908000L
	X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
					      X509_V_FLAG_CRL_CHECK_ALL );
# else	/* OPENSSL_VERSION_NUMBER < 0x00908000L */
	X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
					      X509_V_FLAG_CRL_CHECK_ALL |
					      X509_V_FLAG_ALLOW_PROXY_CERTS );
# endif	/* OPENSSL_VERSION_NUMBER < 0x00908000L */
    } else
	XACML_LOG_FUNC(LOG_DEBUG,
		"%s: Proxy support is disabled, using OpenSSL default verify callback including issuer tests.\n",
		__func__);
#endif	/* DISABLE_PROXY_SUPPORT */


    /* Set max Chain depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);
    SSL_CTX_set_options(ctx, (long)(SSL_OP_ALL|SSL_OP_NO_SSLv2));
    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error setting cipher list (no valid ciphers)\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}


/******************************************************************************
Function:   xacml_io_setup_server_ctx
Description:
    Setting up the SSL server side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *xacml_io_setup_server_ctx(char * cafile,
				   char * cadir,
				   char * servercert,
				   char * serverkey)
{
    SSL_CTX *ctx;

    XACML_LOG_DEBUG(5, "Entering: %s\n", __func__);
    /* ctx = SSL_CTX_new(SSLv23_method()); */

    XACML_LOG_DEBUG(5, "%s: creating context\n", __func__);
    ctx = SSL_CTX_new(SSLv23_server_method());
    if (!ctx)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: No SSL context was created\n", __func__);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Context created\n", __func__);
    }


    if (SSL_CTX_use_PrivateKey_file(ctx, serverkey, SSL_FILETYPE_PEM) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error loading private key from file '%s'\n", __func__, serverkey);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Private Key added to ctx\n", __func__);
    }


    /*
     * Don't set the default OpenSSL verify paths with SSL_CTX_set_default_verify_paths(ctx).
     * Only the configured (Grid) trust anchors are trusted
     */
    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error loading CA file and/or directory '%s'\n", __func__, cafile);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: CA directory added to ctx\n", __func__);
    }

    /* if (SSL_CTX_use_certificate_file(ctx, servercert, SSL_FILETYPE_PEM) != 1) */
    if (SSL_CTX_use_certificate_chain_file(ctx, servercert) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error loading certificate from file '%s'\n", __func__, servercert);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Public certificate chain added to ctx\n", __func__);
    }


    /*
     * Go into verify_callback after default SSL checks
     * Perform signature checks (peer must present certificate, must verify peer
     */


    /* SSL_CTX_set_purpose(ctx, X509_PURPOSE_ANY); */
    SSL_CTX_set_purpose(ctx, X509_PURPOSE_SSL_CLIENT);
/*  X509_STORE_CTX_set_purpose( verify_ctx, X509_PURPOSE_SSL_CLIENT ); */

    /* probably needs to be here */
    ctx->cert_store->check_issued = grid_check_issued_wrapper;
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, scas_verify_callback);
/*  SSL_CTX_set_cert_verify_callback(ctx, scas_gsi_callback_X509_verify_cert, NULL); */

  #if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL );
  #else
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL |
                                                      X509_V_FLAG_ALLOW_PROXY_CERTS );
  #endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */


#if 0 /* Should re-check this flag's placement. Normally found in or before the verify function set to the individual certificate from an SSL handshake. */
   #if OPENSSL_VERSION_NUMBER >= 0x00908000L
     cert->ex_flags |= EXFLAG_PROXY;
   #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */
#endif

    /* Max depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);



/**************************************************************************************/


    /* All (SSLv3, TLS) but not SSLv2 */
    SSL_CTX_set_options(ctx, (long)(SSL_OP_ALL | SSL_OP_NO_SSLv2));

    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error setting cipher list (no valid ciphers)\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    /* Does the Private key match the certificate? */
    if (SSL_CTX_check_private_key(ctx) != 1)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error matching private key\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}




/******************************************************************************
Function:   xacml_io_x509IsCA
Description:
    Tests if the X509 * cert is a CA certificate or not
    => Example from GridSite
Parameters:
    A X509 pointer
Returns:
    0      : Not a CA cert
    1      : This is a CA cert
******************************************************************************/
int xacml_io_x509IsCA (X509 *cert)
{

/*
#define NEW_CA_SEARCH_METHOD
*/
#ifndef NEW_CA_SEARCH_METHOD
    int purpose_id;
    char name[10];

    strncpy(name,"sslclient",10);

    purpose_id = X509_PURPOSE_get_by_sname(name);

    /* final argument to X509_check_purpose() is whether to check for CAness */

    if (X509_check_purpose(cert, purpose_id + X509_PURPOSE_MIN, 1))
        return 1;
    else return 0;
#else
    int                                 element_index = -1;
    int                                 critical;
    BASIC_CONSTRAINTS *                 x509v3_bc = NULL;

    if((x509v3_bc = X509_get_ext_d2i(cert,
                                     NID_basic_constraints,
                                     &critical,
                                     &element_index)) && x509v3_bc->ca)
    {
        /* Free allocated space */
        BASIC_CONSTRAINTS_free(x509v3_bc);

        /* CA cert found */
        return 1;
    }

    /* Not a CA cert */
    return 0;
#endif /* NEW_CA_SEARCH_METHOD */
}




/******************************************************************************
Function:       xacml_io_init_OpenSSL()
Description:    Do the SSL library initialization
Parameters:     void
Returns:
     0 : all is ok
     1 : failure reported must not continue with the SSL library
******************************************************************************/
int xacml_io_init_OpenSSL(void)
{
/*
    if (!THREAD_setup() || !SSL_library_init())
 */

    if (openssl_is_initialized)
        return 0;
    
    /* Note: OpenSSL_add_ssl_algorithms() is macro for SSL_library_init(). Also
     * note that SSL_library_init always returns 1 */
/*    if (!SSL_library_init())
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: ** OpenSSL initialization failed!\n", __func__);
        return 1;
    }*/
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();

    /* Declare OpenSSL to be initialized */
    XACML_LOG_DEBUG(5, "%s: OpenSSL initialized\n", __func__);
    openssl_is_initialized = 1;

    return 0;
}

void xacml_io_cleanup_OpenSSL(void)
{
#ifndef OPENSSL_NO_COMP
    STACK_OF(SSL_COMP)* comp_methods;
#endif

    if (!openssl_is_initialized)    /* nothing to be done (probably) */
	return;

/*    CONF_modules_unload(1);*/
/*    ENGINE_cleanup();*/ /* needs openssl/engine.h */

    /* this is needed to cleanup memory from SSL_CTX_use_certificate_chain_file
     */
    ERR_remove_state(0);

    ERR_free_strings();
    EVP_cleanup();
    CRYPTO_cleanup_all_ex_data();

#ifndef OPENSSL_NO_COMP
    /* See http://www.ultimatepp.org/forum/index.php?t=msg&goto=40347& 
     * for next one, this removes memory from SSL_COMP_get_compression_methods
     * -> load_builtin_compressions */
    comp_methods = SSL_COMP_get_compression_methods();
    if( comp_methods )
	sk_SSL_COMP_free( comp_methods );
#endif /* OPENSSL_NO_COMP */

}


#ifndef DISABLE_PROXY_SUPPORT
/**
 * Returns whether we want to run-time disable proxy support
 */
int xacml_io_get_disable_server_proxy_support(void)   {
    return disable_proxy_support;
}

/**
 * Sets whether we want to run-time disable proxy support
 */
void xacml_io_set_disable_server_proxy_support(int value)   {
    disable_proxy_support=value;
}
#endif /* DISABLE_PROXY_SUPPORT */

/******************************************************************************
Function:       xacml_io_post_connection_check
Description:    Post Connection Checking (only for client)
                DNS (or final /CN=<host>)
                If SubjectAltNames exist, then they must match the host. If they don't exist, then look at the final CN field in the subject.
                All else is a failure
Parameters:     SSL *, char * host
Returns:        long
******************************************************************************/
long xacml_io_post_connection_check(SSL *ssl, const char *host)
{
    X509      *      peer_cert = NULL;
    X509_NAME *      subj;
    char             datad[256];
    int              extcount;
    int              ok = 0;
    int              found_DNS_entry = 0;

    XACML_LOG_DEBUG(5, "%s: entering %s\n", __func__, __func__);

    if (!host)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: error: No hostname supplied to check\n", __func__);
        goto err_occured;
    }

    if (!(peer_cert = SSL_get_peer_certificate(ssl)) || !host)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: error: Could not get peer certificate from the SSL handle for host: %s\n",
                     __func__,
                     host);
        goto err_occured;
    }


  #if OPENSSL_VERSION_NUMBER >= 0x00908000L
    peer_cert->ex_flags |= EXFLAG_PROXY;
  #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */

    /* Compare the subjectAltName DNS value with the host value */
    if ((extcount = X509_get_ext_count(peer_cert)) > 0)
    {
        int i;

        for (i = 0; i < extcount; i++)
        {
            X509_EXTENSION    *ext;
            int NID_from_ext = NID_undef; /* Initialize with undefined NID (Numerical ID of a type of ASN1 object) */

            ext = X509_get_ext(peer_cert, i);
            NID_from_ext = OBJ_obj2nid(X509_EXTENSION_get_object(ext));

            if (NID_from_ext == NID_subject_alt_name)
            {
                int                  j;
/* see line 971 in
 * https://github.com/openssl/openssl/blame/875a644a9047e96dfcce27af876d30460759805e/crypto/asn1/asn1.h
 */
#if (OPENSSL_VERSION_NUMBER < 0x00908000L)
                unsigned char  *data;
#else
                const unsigned char  *data;
#endif
                STACK_OF(CONF_VALUE) *val;
                CONF_VALUE           *nval;
/* prototype of i2v() changed in OpenSSL_0_9_8k, see line 82 in
 * https://github.com/openssl/openssl/blame/babb379849ffb4112792f266f92e9ebb2bd35332/crypto/x509v3/x509v3.h
 */
#if (OPENSSL_VERSION_NUMBER < 0x009080bfL)
                X509V3_EXT_METHOD *meth;
#else
                const X509V3_EXT_METHOD *meth;
#endif
                void                 *ext_str = NULL;

                if (!(meth = X509V3_EXT_get(ext)))
                    break;
                data = ext->value->data;

#if (OPENSSL_VERSION_NUMBER > 0x00907000L)
                if (meth->it)
                    ext_str = ASN1_item_d2i(NULL, &data, ext->value->length,
                                            ASN1_ITEM_ptr(meth->it));
                else
                    ext_str = meth->d2i(NULL, &data, ext->value->length);
#else	/* OPENSSL_VERSION_NUMBER > 0x00907000L */
                ext_str = meth->d2i(NULL, &data, ext->value->length);
#endif	/* OPENSSL_VERSION_NUMBER > 0x00907000L */
                val = meth->i2v(meth, ext_str, NULL);
                for (j = 0;  j < sk_CONF_VALUE_num(val);  j++)
                {
                    nval = sk_CONF_VALUE_value(val, j);
                    if (!strcmp(nval->name, "DNS") && !strcasecmp(nval->value, host))
                    {
                        XACML_LOG_FUNC(LOG_DEBUG, "%s: subjectAltName DNS entry matches host\n", __func__);
                        ok = 1;
                        break;
                    }
                    XACML_LOG_DEBUG(5, "%s: found name: %s value: %s\n", __func__, nval->name, nval->value);
                    if (!strcmp(nval->name, "DNS"))
                    {
                        found_DNS_entry = 1;
                    }
                }

                sk_CONF_VALUE_pop_free( val, X509V3_conf_free );
                if (meth->it)
                    ASN1_item_free((ASN1_VALUE*)ext_str, meth->it);
                else
                    meth->ext_free(ext_str);
                ext_str = NULL;

            }
            if (ok)
                break;
        }
    }


    if (found_DNS_entry && !ok)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: found one or more DNS entries in the subjectAltName but none match the host '%s'.\n", __func__, host);
        goto err_occured;
    }


    /* If the subjectAltName DNS value doesn't match the host value,
     * resort back to the X509v1 solution which is to compare the host value
     * with the final CN value in the cert's subject
     */
    if (!ok && (subj = X509_get_subject_name(peer_cert)) &&
        X509_NAME_get_text_by_NID(subj, NID_commonName, datad, 256) > 0)
    {
        XACML_LOG_FUNC(LOG_DEBUG, "%s: No subjectAltName:DNS entry found, trying final CN field in subject to match contacted host.\n", __func__);
        if (strcasecmp(datad, host) != 0)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: final CN field in subject '%s' and host '%s' do not match!\n", __func__, datad, host);
            goto err_occured;
        }
        else
        {
            XACML_LOG_FUNC(LOG_DEBUG, "%s: final CN field in subject matches host\n", __func__);
            ok = 1;
        }
    }


    /* Clean up */
    if (peer_cert)
    {
        X509_free(peer_cert);
        peer_cert = NULL;
    }

    if (ok)
        return X509_V_OK;
    else
        return X509_V_ERR_APPLICATION_VERIFICATION;

err_occured:
    if (peer_cert)
    {
        X509_free(peer_cert);
        peer_cert = NULL;
    }
    return X509_V_ERR_APPLICATION_VERIFICATION;
}


/******************************************************************************
Function:       xacml_io_x509_chain_to_dn()
Description:    Get the base DN of the X509 proxy cert
Parameters:
                certstack: certificate chain
Returns:        dn (which should be freed)
******************************************************************************/
char * xacml_io_x509_chain_to_dn (STACK_OF(X509) *certstack)
{
    int i = 0;
    int amount_of_CAs = 0;
    int depth = sk_X509_num(certstack);

    /* Do we have a chain at all... */
    if (depth > 0)
    {
        /* Amount of CAs in chain */
        for (i = 0; i < depth; i++)
        {
            if (xacml_io_x509IsCA(sk_X509_value(certstack, i)))
                amount_of_CAs++;
        }

        /* Select the User certificate */
        i = depth - (amount_of_CAs + 1);

        if (i < 0)
        {
            i = 0;
            return NULL;
        }

        /* This returned value must be free'd */
        return X509_NAME_oneline(X509_get_subject_name(sk_X509_value(certstack, i)),NULL,0);
    }
    else
        return NULL;
}


/******************************************************************************
Function:       xacml_io_seed_prng
Description:    seed prng with /dev/urandom
Parameters:     void
Returns:        void
******************************************************************************/
void xacml_io_seed_prng(void)
{
     RAND_load_file("/dev/urandom", 1024L);
}




/****************************************************
Function: xacml_io_print_ssl_error_msg

Variables:
    SSL * ssl   :   pointer to SSL structure (for this connection)
    int ret     :   return code of the previously SSL command
    printmsg    :   0 == don't print the messages, 1 =< print the message

Return:
    0 : Non lethal failure, should consider a retry of the SSL interaction
    1 : Lethal failure, non-recoverable situation
****************************************************/
int xacml_io_print_ssl_error_msg (SSL * ssl, int ret, int printmsg, int * error_code )
{
    int rc = 0;
    int unrecoverable = 0;
    int save_errno;
    unsigned long code;

    if (ssl)
    {
        rc = SSL_get_error (ssl, ret); /* returns int following openssl/ssl.h */
	if (error_code)
	    *error_code = rc;

	switch (rc)
	{
	    case SSL_ERROR_NONE :
	    {
		/* No error */
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: Got error trigger, but there's no error on the stack\n");
		unrecoverable = 0;
		break;
	    }
	    case SSL_ERROR_ZERO_RETURN :
	    {
		/* Shutdown given, possibly not shutdown yet in the underlying
		 * transport layer */
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: Shutdown given, possibly not shutdown yet in the underlying transport layer\n");
		/* Make this unrecoverable. In case the scas client bails out
		 * with an invalid hostname, we seem to get in an infinite loop
		 * here in xacml_io_SSL_recv() */
		unrecoverable = 1;
		break;
	    }
	    case SSL_ERROR_WANT_READ:
	    {
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: operation not completed (wants READ), retry that last I/O operation\n");
		unrecoverable = 0;
		break;
	    }
	    case SSL_ERROR_WANT_WRITE:
	    {
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: operation not completed (wants WRITE), retry that last I/O operation\n");
		unrecoverable = 0;
		break;
	    }
	    case SSL_ERROR_WANT_CONNECT:
	    case SSL_ERROR_WANT_ACCEPT:
	    {
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: operation not completed, underlying I/O not ready, retry that last accept/connect operation. Re-check select() or poll() logics\n");
		unrecoverable = 0;
		break;
	    }
	    case SSL_ERROR_WANT_X509_LOOKUP:
	    {
		if (printmsg)
		    XACML_LOG_FUNC(LOG_DEBUG, "XACML SSL I/O: operation not completed, application callback was called again, should retry.\n");
		unrecoverable = 0;
		break;
	    }
	    case SSL_ERROR_SYSCALL:
	    {
		/* Some I/O error occurred. The OpenSSL error queue may contain
		 * more information on the error. If the error queue is empty
		 * (i.e. ERR_get_error() returns 0), ret can be used to find out
		 * more about the error: If ret == 0, an EOF was observed that
		 * violates the protocol. If ret == -1, the underlying BIO
		 * reported an I/O error (for socket I/O on Unix systems,
		 * consult errno for details). */

		/* unrecoverable failure:
		 * 1. error on stack
		 * 2. no error on stack AND ret==0
		 * 3. no error on stack, ret!=0 and errno is non-zero
		 * hence when there is no error on the stack and errno is zero
		 * we interpret it as a recoverable error. This happens e.g.
		 * when LCAS fails. */
		save_errno=errno;
		if (ERR_peek_error()!=0 || ret==0 || save_errno!=0)
		    unrecoverable=1;
		else
		    unrecoverable=0;

		if (printmsg)
		    XACML_LOG_FUNC(
			(unrecoverable ? LOG_ERR : LOG_INFO),
                        "XACML SSL I/O: Some (%s) I/O error occurred: %s\n",
			(unrecoverable ? "unrecoverable" : "recoverable"),
                        (ret==0 ? "unexpected EOF" :
				(save_errno ? strerror(save_errno)
					    : "unknown reason")));
		break;
	    }
	    case SSL_ERROR_SSL:
	    {
		/* Check ERR_get_error() for more information */
		if (printmsg)
		    XACML_LOG_FUNC(LOG_ERR, "XACML SSL I/O: A failure in the SSL library occurred, usually a protocol error.\n");
		unrecoverable = 1;
		break;
	    }

	    default:
	    {
		if (printmsg)
		    XACML_LOG_FUNC(LOG_ERR, "XACML SSL I/O: Undefined error code 0x%x, not in OpenSSL specs\n",rc);
		unrecoverable = 1;
		break;
	    }
	} /* switch(rc) */
    } /* if (ssl) */
    else
    {
/*        rc = ERR_get_error();*/ /* returns unsigned long following openssl/err.h */
	/* No ssl object and cannot just convert the error from ERR_get_error()
	 * as they don't match at all... */
	XACML_LOG_FUNC(LOG_ERR,"%s: No SSL object\n",__func__);
	if (*error_code)
	    *error_code = SSL_ERROR_SSL; /* No obvious error )-: */
	unrecoverable = 1;
    }

    /* Print error queue */
    if (printmsg)   {
	while ((code=ERR_get_error())!=0)        {
	    XACML_LOG_FUNC(LOG_ERR,"   OpenSSL error queue: %s (function %s from %s)\n",
		    ERR_reason_error_string(code),
		    ERR_func_error_string(code),
		    ERR_lib_error_string(code));
	}
    }

    return unrecoverable;
}


/******************************************************************************
Function:     xacml_io_get_hostname_to_match_client_cert
Description:
              Gets the hostname string to which the peer certificate must match.
              Is this can not be matched automatically, then this feature can be
              used to do so.
Parameters:
              None
Returns:
     NULL : Nothing was set with xacml_io_set_hostname_to_match_client_cert()
    !NULL : String value that must be used in the SSL handshake check to verify
            the peer certificate.
******************************************************************************/
char * xacml_io_get_hostname_to_match_client_cert(void)
{
    return ssl_override_hostname;
}


/******************************************************************************
Function:     xacml_io_set_hostname_to_match_client_cert
Description:
              Sets the hostname string to which the peer certificate must match.
              Is this can not be matched automatically, then this feature can be
              used to do so.
Parameters:
     override_h_n : The string, which must be a hostname, that overrides the
                    hostname used to setup the connection and which will be
                    matched against the peer-certificate information.

Returns:
     0 : Good. The string has been duplicated from 'override_h_n'.
     1 : Bad. Memory allocation has failed.
******************************************************************************/
int xacml_io_set_hostname_to_match_client_cert(char * override_h_n)
{
    if ((ssl_override_hostname = strdup (override_h_n)))
        return 0;
    else
        return 1;
}

/******************************************************************************
Function:     xacml_io_free_hostname_to_match_client_cert
******************************************************************************/
void xacml_io_free_hostname_to_match_client_cert(void)
{
    free(ssl_override_hostname);
    ssl_override_hostname = NULL;
}


/******************************************************************************
 * Static functions
 *****************************************************************************/

/*  When the proxy is located on an NFS mount and on the server side the root
 *  squash option has been enabled, the effective uid is mapped to user 'nobody'
 *  which should not be able to read the proxy file. To work around this
 *  problem, the effective uid of the process is changed to that of the calling
 *  user and once glexec is done, the saved uid is used to restore the identity
 *  of the process,
 */
static int changeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid)
{
    struct passwd * this_pw = NULL;

    /* Seg fault safety */
    if (!real_uid || !saved_uid)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Code error: No storage allocated to the real_uid or saved_uid\n", __func__);
        return 1;
    }

    /* For the moment, this is a global setting that will be used to determine a different account to which the process will need to switch before performing a particular action */
    if (xacml_io_certOwner && strlen(xacml_io_certOwner))
    {
        if ((this_pw = getpwnam(xacml_io_certOwner)))
        {
            *real_uid = this_pw -> pw_uid;
            XACML_LOG_FUNC(LOG_DEBUG, "%s: (Temporarily) changing the UID of the process to \'%s(%d)\'\n", __func__, xacml_io_certOwner, *real_uid);
        }
        else
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: can't lookup the account \'%s\'\n", __func__, xacml_io_certOwner);
            return 1;
        }
    }
    else
    {
        *real_uid = getuid();
    }


    /* Save it */
    *saved_uid = geteuid();
    if (seteuid(*real_uid))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error on changing the process Effective UID with seteuid() to %d\n", __func__, *real_uid);
        return 1;
    }
    return 0;
}

/*  When the proxy is located on an NFS mount and on the server side the root
 *  squash option has been enabled, the effective uid is mapped to user 'nobody'
 *  which should not be able to read the proxy file. To work around this
 *  problem, the effective uid of the process is changed to that of the calling
 *  user and once glexec is done, the saved uid is used to restore the identity
 *  of the process,
 */
static int restoreEffectiveToRealUid (uid_t saved_uid)
{
    if (saved_uid==(uid_t)-1)	{
        XACML_LOG_FUNC(LOG_ERR, "%s: saved_uid is unset, cannot restore\n", __func__);
        return 1;
    }

    /*  Do not forget to put back the original effective uid on the process. */
    if (seteuid(saved_uid))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error on returning seteuid() to the saved UID %d\n", __func__, saved_uid);
        return 1;
    }
    return 0;
}


#ifndef DISABLE_PROXY_SUPPORT
/******************************************************************************
Function:       scas_verify_callback()
Description:    Execute the OpenSSL verify callback
                The ok will be the result of the default verification function
Parameters:     ok, X509_STORE_CTX *
Returns:        int
******************************************************************************/
static int scas_verify_callback(int ok, X509_STORE_CTX *store_ctx)
{
    int             errnum   = X509_STORE_CTX_get_error(store_ctx);
    int             errdepth = X509_STORE_CTX_get_error_depth(store_ctx);
    STACK_OF(X509) *certstack = NULL;
    char *          cert_DN   = NULL;
    char *          issuer_DN = NULL;
    X509           *curr_cert = X509_STORE_CTX_get_current_cert(store_ctx);

    XACML_LOG_DEBUG(3, "%s: - Re-Verify certificate at depth: %i\n", __func__, errdepth);

    cert_DN   = X509_NAME_oneline(X509_get_subject_name(curr_cert), NULL, 0);
    issuer_DN = X509_NAME_oneline(X509_get_issuer_name(curr_cert), NULL, 0);
    XACML_LOG_FUNC(LOG_DEBUG, "%s:   subject  = %s\n", __func__, cert_DN);
    XACML_LOG_FUNC(LOG_DEBUG, "%s:   issuer   = %s\n", __func__, issuer_DN);
    XACML_LOG_FUNC(LOG_DEBUG, "%s:   errnum %d: %s\n", __func__, errnum, X509_verify_cert_error_string(errnum));

    /* Resolve known error state and flag them as 'ok' */
    if (ok != 1)
    {
	/* Have to override X509_V_ERR_INVALID_CA since it can come from
	 * old-style proxies */
        if (errnum == X509_V_ERR_INVALID_CA)
        {
            XACML_LOG_DEBUG(3,  "%s: Override 'Invalid CA' error.\n", __func__);
            ok=1;
        }
        if (errnum == X509_V_ERR_UNABLE_TO_GET_CRL)
        {
            XACML_LOG_DEBUG(3,  "%s: Override 'Unable to get CRL' error\n", __func__);
            ok=1;
        }
        if (errnum == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION)
        {
            XACML_LOG_DEBUG(3,  "%s: Override 'Unhandled Critical Extensions' error. Continue in grid_X509_knownCriticalExts()\n", __func__);
            errnum = grid_X509_knownCriticalExts(curr_cert);
            store_ctx -> error = errnum;
            if (errnum == X509_V_OK)
            {
                XACML_LOG_DEBUG(3,  "%s:   grid_X509_knownCriticalExts() returned X509_V_OK.\n", __func__);
                ok=1;
            }
            else
            {
                XACML_LOG_DEBUG(3,  "%s:   failure: grid_X509_knownCriticalExts() returned: %d\n", __func__, errnum);
            }
        }
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
        /* I don't want to do this, really, but I have yet to figure out
           how to get openssl 0.9.8 to accept proxy certificates when
           opening a new SSL connection...
        */
        if (errnum == X509_V_ERR_INVALID_PURPOSE)
        {
            XACML_LOG_DEBUG(3,  "%s: Override 'Unsupported certificate purpose' error\n", __func__);
            ok=1;
        }
#endif	/* OPENSSL_VERSION_NUMBER >= 0x00908000L */

#if 0
        if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ok=1;
        /* since we are just checking the certificates, it is
         * ok if they are self signed. But we should still warn
         * the user.
         */
        if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
        /* Continue after extension errors too */
        if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ok=1;
        if (ctx->error == X509_V_ERR_INVALID_PURPOSE) ok=1;
        if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
        if (ctx->error == X509_V_ERR_CRL_HAS_EXPIRED) ok=1;
        if (ctx->error == X509_V_ERR_CRL_NOT_YET_VALID) ok=1;
#endif

        if (ok != 1)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Verification failure at depth %d\n", __func__, errdepth);
            XACML_LOG_DEBUG(3, "%s:     Failure still applies, no override performed for this error\n", __func__);
            XACML_LOG_FUNC(LOG_NOTICE, "%s:      issuer   = %s\n", __func__, issuer_DN);
            XACML_LOG_FUNC(LOG_NOTICE, "%s:      subject  = %s\n", __func__, cert_DN);
        }
    }

    /*
     * We've now got the last certificate - the identity being used for
     * this connection. At this point we check the whole chain for valid
     * CAs or, failing that, GSI-proxy validity using grid_verifyProxy
     */
    if ( (errdepth == 0) && (ok == 1) )
    {
        certstack = (STACK_OF(X509) *) X509_STORE_CTX_get1_chain(store_ctx);

        XACML_LOG_DEBUG(5, "%s: Final certificate in the stack reached, starting full chain test in grid_verifyProxy(certstack)\n", __func__);
        errnum = grid_verifyProxy(certstack);
        store_ctx->error = errnum;
        if (errnum == X509_V_OK) ok=1;

	/* Revived this pop_free - Brian: Nov 12 2010 */
         /* Originally, this memory leak was required because of a later double-free */
        sk_X509_pop_free( certstack, X509_free );
        certstack = NULL;
    }


    if (ok != 1)
    {
        XACML_LOG_FUNC(LOG_NOTICE, "%s: error message = %s\n", 
                 __func__, X509_verify_cert_error_string (errnum));
    }


    /* The current way of working needs freeing */
    free (cert_DN);
    free (issuer_DN);

    return ok;
}


/******************************************************************************
Function:       grid_X509_knownCriticalExts
Description:    Check if the Critical Extention known to proxy certificates
                are set and understood.
Parameters:     X509 * cert
Returns:        unsigned long : X509_V_OK (good), X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION (bad)
******************************************************************************/
#define PROXYCERTINFO_OID      "1.3.6.1.5.5.7.1.14"
#define OLD_PROXYCERTINFO_OID  "1.3.6.1.4.1.3536.1.222"

static int grid_X509_knownCriticalExts(X509 *cert)
{
   int  i = 0;
   char critical[80];
   X509_EXTENSION *ex;
   char            data[256];


   XACML_LOG_DEBUG(5, "Entering %s\n", __func__);
   for (i = 0; i < X509_get_ext_count(cert); ++i)
   {
        X509_NAME_oneline(X509_get_issuer_name(cert), data, 256);
        XACML_LOG_FUNC(LOG_DEBUG, "%s:  issuer  = %s\n", __func__, data);
        X509_NAME_oneline(X509_get_subject_name(cert), data, 256);
        XACML_LOG_FUNC(LOG_DEBUG, "%s:  subject = %s\n", __func__, data);
        ex = X509_get_ext(cert, i);

        if (X509_EXTENSION_get_critical(ex) &&
                                 !X509_supported_extension(ex))
        {
            OBJ_obj2txt(critical, sizeof(critical), X509_EXTENSION_get_object(ex), 1);

            XACML_LOG_FUNC(LOG_DEBUG, "%s: Critical extension found: %s", __func__, critical);
            if (strcmp(critical, PROXYCERTINFO_OID) == 0)     return X509_V_OK;
            if (strcmp(critical, OLD_PROXYCERTINFO_OID) == 0) return X509_V_OK;

            return X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION;
        }
   }
   return X509_V_OK;
}


/******************************************************************************
Function:   grid_verifyProxy
Description:
    Tries to verify the proxies in the certstack
Note: X509_V_ERR_* and X509_V_OK are #define-s in openssl/x509_vfy.h. In the
    same header file, the error field in a X509_STORE_CTX is an int. So
    presumable we can assume it is an int.
******************************************************************************/
static int grid_verifyProxy( STACK_OF(X509) *certstack )
{
    int      i = 0;
    X509    *cert = NULL;
    time_t   now = time((time_t *)NULL);
    size_t   len = 0;             /* Lengths of issuer and cert DN */
    size_t   len2 = 0;            /* Lengths of issuer and cert DN */
    int      is_old_style_proxy = 0;
    int      is_limited_proxy = 0;
    int      prevIsLimited = 0;   /* previous cert was proxy and limited */
    char     cert_DN   [256];     /* Pointer to current-certificate-in-certstack's DN */
    char     issuer_DN [256];     /* Pointer to issuer-of-current-cert-in-certstack's DN */
    char    *proxy_part_DN = NULL;
    int      depth = sk_X509_num (certstack);
    int      amount_of_CAs = 0;
    ASN1_INTEGER * cert_Serial,*issuer_Serial;


    XACML_LOG_DEBUG(5, "%s: --- Welcome to the grid_verifyProxy function ---\n", __func__);

    if (depth == 0)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Empty certificate chain presented!\n", __func__);
        return X509_V_ERR_APPLICATION_VERIFICATION;
    }


    /* And there was (current) time... */
    time(&now);

    /* How many CA certs are there in the certstack? */
    for (i = 0; i < depth; i++)
    {
        if (xacml_io_x509IsCA(sk_X509_value(certstack, i)))
            amount_of_CAs++;
    }


    XACML_LOG_DEBUG(5, "%s: #CA's = %d , depth = %d\n", __func__, amount_of_CAs, depth);

    /* Check if our certificate chain could hold anything useful, like proxies or EECs */
    /* Need to change */
    if ((amount_of_CAs + 2) > depth)
    {
        if ((depth - amount_of_CAs) > 0)
        {
            XACML_LOG_FUNC(LOG_DEBUG, "%s: No proxy certificate in certificate stack to check.\n", __func__);
            return X509_V_OK;
        }
        else
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: No personal certificate (neither proxy or user certificate) found in the certficiate stack.", __func__);
            return X509_V_ERR_APPLICATION_VERIFICATION;
        }
    }


    /*
     * Changed this value to start checking the proxy and such and
     * to skip the CA and the user_cert
     */

    for (i = depth - (amount_of_CAs + 2); i >= 0; i--)
    {
        /* Check for X509 certificate and point to it with 'cert' */
        if ( (cert = sk_X509_value(certstack, i)) != NULL )
        {
            X509_NAME_oneline(X509_get_subject_name(cert), cert_DN,   256);
            X509_NAME_oneline(X509_get_issuer_name(cert),  issuer_DN, 256);
            len       = strlen( cert_DN );
            len2      = strlen( issuer_DN );

            XACML_LOG_DEBUG(5, "%s: Proxy to verify:\n", __func__ );
            XACML_LOG_DEBUG(5, "%s:   Issuer DN: %s\n", __func__, issuer_DN );
            XACML_LOG_DEBUG(5, "%s:   DN:        %s\n", __func__, cert_DN );

            if (now < grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notBefore(cert))))
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Proxy certificate is not yet valid.\n", __func__);
                return X509_V_ERR_CERT_NOT_YET_VALID;
            }

            if (now > grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notAfter(cert))))
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Proxy certificate expired.\n", __func__);
                return X509_V_ERR_CERT_HAS_EXPIRED;
            }

            /* User not allowed to sign shortened DN */
            /* Note: Only true for GT2 proxies */
            if (len2 > len)
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: It is not allowed to sign a shorthened DN.\n", __func__);
                return X509_V_ERR_INVALID_CA;
            }

            /* Proxy subject must begin with issuer. */
            if (strncmp(cert_DN, issuer_DN, len2) != 0)
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Proxy subject must begin with the issuer.\n", __func__);
                return X509_V_ERR_INVALID_CA;
            }

            /* Set pointer to end of base DN in cert_DN */
            proxy_part_DN = &cert_DN[len2];

            /* First attempt at support for Old and New style GSI
               proxies: /CN=anything is ok for now */
            if (strncmp(proxy_part_DN, "/CN=", (size_t)4) != 0)
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Could not find a /CN= structure in the DN, thus it is not a proxy.\n", __func__);
                return X509_V_ERR_APPLICATION_VERIFICATION;
            }

            if (strncmp(proxy_part_DN, "/CN=proxy", (size_t)9) == 0)
            {
                XACML_LOG_FUNC(LOG_DEBUG, "%s: Current certificate is an old style proxy.\n",__func__);
                is_old_style_proxy = 1;
                is_limited_proxy = 0;
            }
            else if (strncmp(proxy_part_DN, "/CN=limited proxy", (size_t)17) == 0)
            {
                XACML_LOG_FUNC(LOG_DEBUG, "%s: Current certificate is an old limited style proxy.\n",__func__);
                is_old_style_proxy = 1;
                is_limited_proxy = 1;
            }
            else
            {
                XACML_LOG_FUNC(LOG_DEBUG, "Current certificate is a GSI/RFC3820 proxy.\n",__func__);
            }

            if ( is_old_style_proxy )
            {
                cert_Serial = X509_get_serialNumber( cert );
		grid_get_serialStr(cert_Serial);
                issuer_Serial = X509_get_serialNumber( sk_X509_value(certstack, i+1));
		grid_get_serialStr(issuer_Serial);
                if (cert_Serial && issuer_Serial)
                {
                    if ( ASN1_INTEGER_cmp( cert_Serial, issuer_Serial ) )
                    {
                        XACML_LOG_FUNC(LOG_WARNING, "%s: Serial numbers do not match.\n",__func__ );
                    }
                }
            }

            if ( is_limited_proxy )
            {
                prevIsLimited = 1;
                if (i > 0) XACML_LOG_FUNC(LOG_DEBUG, "%s: Found limited proxy.\n",__func__);
            }
            else
            {
                if (prevIsLimited)
                {
                    XACML_LOG_FUNC(LOG_ERR, "%s: Proxy chain integrity error. Previous proxy in chain was limited, but this one is a regular proxy.\n",__func__);
                    return X509_V_ERR_APPLICATION_VERIFICATION;
                }
            }

            XACML_LOG_DEBUG(5, "%s:   Proxy is valid\n", __func__);
        }
    }

    return X509_V_OK;
}

/**
 * Note that timegm() is non-standard. Linux manpage advices the following
 * substition instead.
 */
static time_t my_timegm(struct tm *tm)
{
   time_t ret;
   char *tz;

   tz = getenv("TZ");
   setenv("TZ", "", 1);
   tzset();
   ret = mktime(tm);
   if (tz)
       setenv("TZ", tz, 1);
   else
       unsetenv("TZ");
   tzset();

   return ret;
}



/**
 *  (Use ASN1_STRING_data() to convert ASN1_GENERALIZEDTIME to char * if
 *   necessary)
 */

static time_t grid_asn1TimeToTimeT(char *asn1time)
{
   char   zone;
   struct tm time_tm;
   size_t len;

   /* Make sure to start with an empty struct */
   memset(&time_tm,0,sizeof(struct tm));

   len = strlen(asn1time);

   if ((len != 13) && (len != 15)) return 0; /* dont understand */

   if ((len == 13) &&
       ((sscanf(asn1time, "%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   if ((len == 15) &&
       ((sscanf(asn1time, "20%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   /* time format fixups */

   if (time_tm.tm_year < 90) time_tm.tm_year += 100;
   --(time_tm.tm_mon);

   return my_timegm(&time_tm);
}


static int grid_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer)
/* We change the default callback to use our wrapper and discard errors
   due to GSI proxy chains (ie where users certs act as CAs) */
{
    int ret;
    char * cert_DN      = NULL;
    char * cert_issuer  = NULL;
    char * issuer_DN    = NULL;

    XACML_LOG_DEBUG(5, "Starting %s\n", __func__);

    cert_DN     = X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
    cert_issuer = X509_NAME_oneline(X509_get_issuer_name(x), NULL, 0);
    issuer_DN   = X509_NAME_oneline(X509_get_subject_name(issuer), NULL, 0);

    XACML_LOG_DEBUG(3, "%s:  subject for cert   = %s\n", __func__, cert_DN);
    XACML_LOG_DEBUG(3, "%s:  issuer  for cert   = %s\n", __func__, cert_issuer);
    XACML_LOG_DEBUG(3, "%s:  subject for issuer = %s\n", __func__, issuer_DN);

    free(cert_DN);
    free(cert_issuer);
    free(issuer_DN);

    /* If all is ok, move out of here */
    ret = X509_check_issued(issuer, x);
    if (ret == X509_V_OK)
    {
        XACML_LOG_FUNC(LOG_DEBUG, "%s:  X509_check_issued() == ok\n", __func__);
        return 1;
    }

    XACML_LOG_FUNC(LOG_DEBUG, "%s: check_issued returned: %s\n", __func__, X509_verify_cert_error_string (ret));

    /* Non self-signed certs without signing are ok if they passed
           the other checks inside X509_check_issued. Is this enough? */
    if ((ret == X509_V_ERR_KEYUSAGE_NO_CERTSIGN) &&
        (X509_subject_name_cmp(issuer, x) != 0)) return 1;

    /* If we haven't asked for issuer errors don't set ctx */
#if OPENSSL_VERSION_NUMBER < 0x00908000L
    if (!(ctx->flags & X509_V_FLAG_CB_ISSUER_CHECK))
#else
    if (!(ctx->param->flags & X509_V_FLAG_CB_ISSUER_CHECK))
#endif
    {
        XACML_LOG_FUNC(LOG_DEBUG, "%s: not asked for issuer errors to be set, returning with X509_V_OK\n", __func__);
        return X509_V_OK;
    }

    ctx->error = ret;
    ctx->current_cert = x;
    ctx->current_issuer = issuer;
    ret = ctx->verify_cb(0, ctx);

    XACML_LOG_FUNC(LOG_DEBUG, "%s: ctx->verify_cb(0, ctx) == %d\n", __func__,ret);
    return ret;
}


/******************************************************************************
Function:     scas_client_print_serial
Description:
              Print a certificate serial number in hex (e.g. 088E)
Parameters:
              ASN1_INTEGER, the output of X509_get_serialNumber()
******************************************************************************/
static void grid_get_serialStr(ASN1_INTEGER *serial) {
    unsigned char *temp;
    unsigned char serialNumberDER[127];
    static unsigned char serialStr[SERIALSTR_LENGTH];
    int serialLen, j = 0;
    int ret;

    /* Get length and fill serialNumberDER */
    temp=serialNumberDER;
    serialLen = i2c_ASN1_INTEGER( serial, &temp );

    /* Clear serialStr buffer */
    memset(serialStr,'\0',sizeof( serialStr));

    temp = serialStr;
    for (j = 0; j < serialLen; j++)
    {
	ret=snprintf( (char*)temp, SERIALSTR_LENGTH, "%02X", serialNumberDER[j] );
	if (ret<0)	{
	    XACML_LOG_FUNC(LOG_ERR, "%s: Error converting serialNumberDER: %s", __func__, strerror(errno));
	    return;
	}
	/* Note: ret>=0 by now */
	if ((size_t)ret>=SERIALSTR_LENGTH)	{
	    XACML_LOG_FUNC(LOG_WARNING, "%s: Error converting serialNumberDER, buffer too small", __func__);
	    return;
	}

	temp += 2;
    }
    
    XACML_LOG_FUNC(LOG_DEBUG, "%s: Serial number: %s\n", __func__,serialStr);

}

#endif /* DISABLE_PROXY_SUPPORT */
